#include "linda.h"

static struct _TS_LIST _linda_tuplespace[SIZEOF_TS];

static pthread_t tids[MAX_NO_OF_THREADS];
static int tid_idx;

void linda_init(void)
{
   int i, j;
   
   for (i=0; i<SIZEOF_TS; i++) {
       pthread_mutex_init(&_linda_tuplespace[i].lock, NULL);
       _linda_tuplespace[i].anode = _TS_NODE_NULL;
   }
   tid_idx = 0;
} 

void linda_end(void)
{
   int i;
   void *st;

   for (i=0; i<tid_idx; i++)
       pthread_join(tids[i], &st);
}

typedef void *(*PTHREAD_FTYPE)(void *);

void spawn (void (*f)(void))
{
   PTHREAD_FTYPE fi;
   if (tid_idx >= MAX_NO_OF_THREADS) {
      fprintf(stderr, "Maximum number of threads exceeded!\n");
      exit(-1);
   }

   fi = (PTHREAD_FTYPE)f;
   pthread_create(&(tids[tid_idx++]), (pthread_attr_t *)0, fi, (void *)0);
}

int match(struct _INODE_LIST *inode1, struct _INODE_LIST *inode2)
{
    struct _INODE_LIST *ptr1, *ptr2;

    if ((inode1 == NULL) || (inode2 == NULL)) return 0;

    for (ptr1 = inode1, ptr2 = inode2; ((ptr1 != NULL) && (ptr2 != NULL));
         ptr1 = ptr1->next, ptr2 = ptr2->next) {

        if (ptr1->field_tag == D_FIELD) {
           if (ptr2->field_tag == D_FIELD) {
              switch (ptr1->type_tag) {
               case CHAR_TYPE:
                 if ((ptr2->type_tag != CHAR_TYPE) ||
                     (ptr1->node.c != ptr2->node.c)) return 0;
                 break;
               case CHARP_TYPE:
                 if ((ptr2->type_tag != CHARP_TYPE) ||
                     strcmp(ptr1->node.s, ptr2->node.s)) return 0;
                 break;
               case INT_TYPE:
                 if ((ptr2->type_tag != INT_TYPE) ||
                     (ptr1->node.i != ptr2->node.i)) return 0;
                 break;
               case FLOAT_TYPE:
                 if ((ptr2->type_tag != FLOAT_TYPE) ||
                     (ptr1->node.f != ptr2->node.f)) return 0;
                 break;
               default:
                 fprintf(stderr, "Shit!\n");
                 exit(-1);
                 break;
              }
           } else { /* ptr2->field_tag == Q_FIELD */
              switch (ptr1->type_tag) {
               case CHAR_TYPE:
                 if (ptr2->type_tag != CHAR_TYPE) return 0;
                 break;
               case CHARP_TYPE:
                 if (ptr2->type_tag != CHARP_TYPE) return 0;
                 break;
               case INT_TYPE:
                 if (ptr2->type_tag != INT_TYPE) return 0;
                 break;
               case FLOAT_TYPE:
                 if (ptr2->type_tag != FLOAT_TYPE) return 0;
                 break;
               default:
                 fprintf(stderr, "Shit!\n");
                 exit(-1);
                 break;
              }
           }
        } else { /* Q_FIELD */
           if (ptr2->field_tag == D_FIELD) {
              switch (ptr1->type_tag) {
               case CHAR_TYPE:
                 if (ptr2->type_tag != CHAR_TYPE) return 0;
                 break;
               case CHARP_TYPE:
                 if (ptr2->type_tag != CHARP_TYPE) return 0;
                 break;
               case INT_TYPE:
                 if (ptr2->type_tag != INT_TYPE) return 0;
                 break;
               case FLOAT_TYPE:
                 if (ptr2->type_tag != FLOAT_TYPE) return 0;
                 break;
               default:
                 fprintf(stderr, "Shit!\n");
                 exit(-1);
                 break;
              }
           } else {
             return 0; /* Q_FIELD shouldn't match another Q_FIELD */
           }
        }
    }

    if ((ptr1 == NULL) && (ptr2 == NULL)) return 1;
    else return 0;
}

void bind_val(struct _INODE_LIST *inode1, struct _INODE_LIST *inode2)
{
    struct _INODE_LIST *ptr1, *ptr2;

    if ((inode1 == NULL) || (inode2 == NULL)) return;

    for (ptr1 = inode1, ptr2 = inode2; (ptr1 != NULL) && (ptr2 != NULL);
         ptr1 = ptr1->next, ptr2 = ptr2->next) {

        if (ptr1->field_tag == D_FIELD) {
           if (ptr2->field_tag == Q_FIELD) {
              switch (ptr1->type_tag) {
               case CHAR_TYPE:
                 *ptr2->node.cp = ptr1->node.c;
                 break;
               case CHARP_TYPE:
                 strcpy(ptr2->node.s, ptr1->node.s);
                 break;
               case INT_TYPE:
                 *ptr2->node.ip = ptr1->node.i;
                 break;
               case FLOAT_TYPE:
                 *(ptr2->node.fp) = ptr1->node.f;
                 break;
               default:
                 fprintf(stderr, "Shit!\n");
                 exit(-1);
                 break;
              }
           } 
        } else { /* Q_FIELD */
           if (ptr2->field_tag == D_FIELD) {
              switch (ptr1->type_tag) {
               case CHAR_TYPE:
                 *ptr1->node.cp = ptr2->node.c;
                 break;
               case CHARP_TYPE:
                 strcpy(ptr1->node.s, ptr2->node.s);
                 break;
               case INT_TYPE:
                 *ptr1->node.ip = ptr2->node.i;
                 break;
               case FLOAT_TYPE:
                 *ptr1->node.fp = ptr2->node.f;
                 break;
               default:
                 fprintf(stderr, "Shit!\n");
                 exit(-1);
                 break;
              }
           }
        }
    }
}

void out (char *tuple_mask, ...) 
{ 
    va_list tuple_list; 
    char *mask_ptr; 
    int len, hash, i;
    struct _INODE_LIST *inode, *node_list, *node_ptr;
    struct _TS_NODE *tsnode, *tsptr;
 
    len = (strlen(tuple_mask) / 2);
    hash = 0;

    tsnode = (struct _TS_NODE *)(malloc(sizeof(struct _TS_NODE)));
    tsnode->type = OUT;
    tsnode->inode = NULL;
    pthread_cond_init (&(tsnode->cv), NULL);

    node_ptr = NULL;

    va_start(tuple_list, tuple_mask); 

    for (mask_ptr = tuple_mask, i = 1; *mask_ptr; mask_ptr+=2, i++) 
    { 
     if (*mask_ptr == '%') {
         switch (mask_ptr[1]) 
         { 
         case 's': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.s = va_arg(tuple_list, char *); 
           inode->type_tag = CHARP_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 3 * i;
           break;
         case 'd': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.i = va_arg(tuple_list, int); 
           inode->type_tag = INT_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 5 * i;
           break;
         case 'f': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.f = va_arg(tuple_list, float); 
           inode->type_tag = FLOAT_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 7 * i;
           break;
         case 'c': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.c = va_arg(tuple_list, char); 
           inode->type_tag = CHAR_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 11 * i;
           break;
         default:
           fprintf(stderr, "Unknown type of an out.\n");
           exit(-1);
         } 
     } else {
        if (*mask_ptr == '?') {
           fprintf(stderr, "Cannot do a '?' on a out!");
           exit(-1);
        }
     }

     if (tsnode->inode == NULL) {
        tsnode->inode = node_ptr = inode;
     } else {
       node_ptr->next = inode;
       node_ptr = inode;
     }
    } 

    va_end(tuple_list); 

    /* Enter it into the tuple space */

    len = (len + hash) % SIZEOF_TS;
    pthread_mutex_lock(&_linda_tuplespace[len].lock);
    
    for (tsptr = _linda_tuplespace[len].anode; tsptr != NULL; 
         tsptr = tsptr->next) {

         if (match(tsnode->inode, tsptr->inode)) {
            if (tsptr->type == IN) {
               bind_val(tsnode->inode, tsptr->inode);
               pthread_cond_signal (&(tsptr->cv));
               goto GET_OUT;
            } else {
              if (tsptr->type == RD) {
                 bind_val(tsnode->inode, tsptr->inode);
                 pthread_cond_signal (&(tsptr->cv));
              }
            }
         }
    }

    tsnode->next = _linda_tuplespace[len].anode;
    _linda_tuplespace[len].anode = tsnode;

GET_OUT:

    pthread_mutex_unlock(&_linda_tuplespace[len].lock);
}

void in (char *tuple_mask, ...) 
{ 
    va_list tuple_list; 
    char *mask_ptr; 
    int len, hash, i;
    struct _INODE_LIST *inode, *node_list, *node_ptr;
    struct _TS_NODE *tsnode, *ptsptr, *tsptr;
 
    len = (strlen(tuple_mask) / 2);
    hash = 0;
    
    tsnode = (struct _TS_NODE *)(malloc(sizeof(struct _TS_NODE)));
    tsnode->type = IN;
    tsnode->inode = NULL;
    pthread_cond_init (&(tsnode->cv), NULL);

    node_ptr = NULL;

    va_start(tuple_list, tuple_mask); 

    for (mask_ptr = tuple_mask, i = 1; *mask_ptr; mask_ptr+=2, i++) 
    { 
     if (*mask_ptr == '%') {
         switch (mask_ptr[1]) 
         { 
         case 's': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.s = va_arg(tuple_list, char *); 
           inode->type_tag = CHARP_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 3 * i;
           break;
         case 'd': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.i = va_arg(tuple_list, int); 
           inode->type_tag = INT_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 5 * i;
           break;
         case 'f': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.f = va_arg(tuple_list, float); 
           inode->type_tag = FLOAT_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 7 * i;
           break;
         case 'c': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.c = va_arg(tuple_list, char); 
           inode->type_tag = CHAR_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 11 * i;
           break;
         default:
           fprintf(stderr, "Unknown type of an out.\n");
           exit(-1);
         } 
     } else {
        if (*mask_ptr == '?') {
           switch (mask_ptr[1]) 
           { 
            case 's': 
              inode = 
                (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
              inode->node.s = va_arg(tuple_list, char *); 
              inode->type_tag = CHARP_TYPE;
              inode->field_tag = Q_FIELD;
              inode->next = NULL;
              hash += 3 * i;
              break;
            case 'd': 
              inode = 
                (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
              inode->node.ip = va_arg(tuple_list, int *); 
              inode->type_tag = INT_TYPE;
              inode->field_tag = Q_FIELD;
              inode->next = NULL;
              hash += 5 * i;
              break;
            case 'f': 
              inode = 
                (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
              inode->node.fp = va_arg(tuple_list, float *); 
              inode->type_tag = FLOAT_TYPE;
              inode->field_tag = Q_FIELD;
              inode->next = NULL;
              hash += 7 * i;
              break;
            case 'c': 
              inode = 
                (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
              inode->node.cp = va_arg(tuple_list, char *); 
              inode->type_tag = CHAR_TYPE;
              inode->field_tag = Q_FIELD;
              inode->next = NULL;
              hash += 11 * i;
              break;
            default:
              fprintf(stderr, "Unknown type of an in.\n");
              exit(-1);
         }
        } else {
           fprintf(stderr, "Don't know what it is - in ?!");
           exit(-1);
        }
     }

     if (tsnode->inode == NULL) {
        tsnode->inode = node_ptr = inode;
     } else {
       node_ptr->next = inode;
       node_ptr = inode;
     }
    } 

    va_end(tuple_list); 

    /* Attempt to find it in the tuple space */

    len = (len + hash) % SIZEOF_TS;
    pthread_mutex_lock(&_linda_tuplespace[len].lock);
    
    ptsptr = NULL;

    for (tsptr = _linda_tuplespace[len].anode; tsptr != NULL; 
         tsptr = tsptr->next) {

         if (match(tsnode->inode, tsptr->inode)) {
            if (tsptr->type == OUT) {
               bind_val(tsnode->inode, tsptr->inode);

               /* Delink the out tuple */
               if (ptsptr == NULL) {
                  _linda_tuplespace[len].anode = tsptr->next;
               } else {
                  ptsptr->next = tsptr->next;
               }

               goto GET_OUT;
            } 
         }

         ptsptr = tsptr;
    }

    tsnode->next = _linda_tuplespace[len].anode;
    _linda_tuplespace[len].anode = tsnode;
    pthread_cond_wait (&(tsnode->cv), &(_linda_tuplespace[len].lock));
    pthread_cond_destroy(&(tsnode->cv));
GET_OUT:

    pthread_mutex_unlock(&_linda_tuplespace[len].lock);
}

void rd (char *tuple_mask, ...) 
{ 
    va_list tuple_list; 
    char *mask_ptr; 
    int len, hash, i;
    struct _INODE_LIST *inode, *node_list, *node_ptr;
    struct _TS_NODE *tsnode, *ptsptr, *tsptr;
 
    len = (strlen(tuple_mask) / 2);
    hash = 0;
    
    tsnode = (struct _TS_NODE *)(malloc(sizeof(struct _TS_NODE)));
    tsnode->type = RD;
    tsnode->inode = NULL;
    pthread_cond_init (&(tsnode->cv), NULL);

    node_ptr = NULL;

    va_start(tuple_list, tuple_mask); 

    for (mask_ptr = tuple_mask, i = 1; *mask_ptr; mask_ptr+=2, i++) 
    { 
     if (*mask_ptr == '%') {
         switch (mask_ptr[1]) 
         { 
         case 's': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.s = va_arg(tuple_list, char *); 
           inode->type_tag = CHARP_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 3 * i;
           break;
         case 'd': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.i = va_arg(tuple_list, int); 
           inode->type_tag = INT_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 5 * i;
           break;
         case 'f': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.f = va_arg(tuple_list, float); 
           inode->type_tag = FLOAT_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 7 * i;
           break;
         case 'c': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.c = va_arg(tuple_list, char); 
           inode->type_tag = CHAR_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 11 * i;
           break;
         default:
           fprintf(stderr, "Unknown type of an out.\n");
           exit(-1);
         } 
     } else {
        if (*mask_ptr == '?') {
           switch (mask_ptr[1]) 
           { 
            case 's': 
              inode = 
                (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
              inode->node.s = va_arg(tuple_list, char *); 
              inode->type_tag = CHARP_TYPE;
              inode->field_tag = Q_FIELD;
              inode->next = NULL;
              hash += 3 * i;
              break;
            case 'd': 
              inode = 
                (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
              inode->node.ip = va_arg(tuple_list, int *); 
              inode->type_tag = INT_TYPE;
              inode->field_tag = Q_FIELD;
              inode->next = NULL;
              hash += 5 * i;
              break;
            case 'f': 
              inode = 
                (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
              inode->node.fp = va_arg(tuple_list, float *); 
              inode->type_tag = FLOAT_TYPE;
              inode->field_tag = Q_FIELD;
              inode->next = NULL;
              hash += 7 * i;
              break;
            case 'c': 
              inode = 
                (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
              inode->node.cp = va_arg(tuple_list, char *); 
              inode->type_tag = CHAR_TYPE;
              inode->field_tag = Q_FIELD;
              inode->next = NULL;
              hash += 11 * i;
              break;
            default:
              fprintf(stderr, "Unknown type of an in.\n");
              exit(-1);
         }
        } else {
           fprintf(stderr, "Don't know what it is - in ?!");
           exit(-1);
        }
     }

     if (tsnode->inode == NULL) {
        tsnode->inode = node_ptr = inode;
     } else {
       node_ptr->next = inode;
       node_ptr = inode;
     }
    } 

    va_end(tuple_list); 

    /* Attempt to find it in the tuple space */

    len = (len + hash) % SIZEOF_TS;
    pthread_mutex_lock(&_linda_tuplespace[len].lock);
    
    ptsptr = NULL;

    for (tsptr = _linda_tuplespace[len].anode; tsptr != NULL; 
         tsptr = tsptr->next) {

         if (match(tsnode->inode, tsptr->inode)) {
            if (tsptr->type == OUT) {
               bind_val(tsnode->inode, tsptr->inode);

               /* Don't delink the out tuple */

               goto GET_OUT;
            } 
         }

         ptsptr = tsptr;
    }

    tsnode->next = _linda_tuplespace[len].anode;
    _linda_tuplespace[len].anode = tsnode;
    pthread_cond_wait (&(tsnode->cv), &(_linda_tuplespace[len].lock));
    pthread_cond_destroy(&(tsnode->cv));
GET_OUT:

    pthread_mutex_unlock(&_linda_tuplespace[len].lock);
}

int inp (char *tuple_mask, ...) 
{ 
    va_list tuple_list; 
    char *mask_ptr; 
    int len, hash, i;
    struct _INODE_LIST *inode, *node_list, *node_ptr;
    struct _TS_NODE *tsnode, *ptsptr, *tsptr;
 
    len = (strlen(tuple_mask) / 2);
    hash = 0;
    
    tsnode = (struct _TS_NODE *)(malloc(sizeof(struct _TS_NODE)));
    tsnode->type = IN;
    tsnode->inode = NULL;

    node_ptr = NULL;

    va_start(tuple_list, tuple_mask); 

    for (mask_ptr = tuple_mask, i = 1; *mask_ptr; mask_ptr+=2, i++) 
    { 
     if (*mask_ptr == '%') {
         switch (mask_ptr[1]) 
         { 
         case 's': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.s = va_arg(tuple_list, char *); 
           inode->type_tag = CHARP_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 3 * i;
           break;
         case 'd': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.i = va_arg(tuple_list, int); 
           inode->type_tag = INT_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 5 * i;
           break;
         case 'f': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.f = va_arg(tuple_list, float); 
           inode->type_tag = FLOAT_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 7 * i;
           break;
         case 'c': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.c = va_arg(tuple_list, char); 
           inode->type_tag = CHAR_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 11 * i;
           break;
         default:
           fprintf(stderr, "Unknown type of an out.\n");
           exit(-1);
         } 
     } else {
        if (*mask_ptr == '?') {
           switch (mask_ptr[1]) 
           { 
            case 's': 
              inode = 
                (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
              inode->node.s = va_arg(tuple_list, char *); 
              inode->type_tag = CHARP_TYPE;
              inode->field_tag = Q_FIELD;
              inode->next = NULL;
              hash += 3 * i;
              break;
            case 'd': 
              inode = 
                (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
              inode->node.ip = va_arg(tuple_list, int *); 
              inode->type_tag = INT_TYPE;
              inode->field_tag = Q_FIELD;
              inode->next = NULL;
              hash += 5 * i;
              break;
            case 'f': 
              inode = 
                (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
              inode->node.fp = va_arg(tuple_list, float *); 
              inode->type_tag = FLOAT_TYPE;
              inode->field_tag = Q_FIELD;
              inode->next = NULL;
              hash += 7 * i;
              break;
            case 'c': 
              inode = 
                (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
              inode->node.cp = va_arg(tuple_list, char *); 
              inode->type_tag = CHAR_TYPE;
              inode->field_tag = Q_FIELD;
              inode->next = NULL;
              hash += 11 * i;
              break;
            default:
              fprintf(stderr, "Unknown type of an in.\n");
              exit(-1);
         }
        } else {
           fprintf(stderr, "Don't know what it is - in ?!");
           exit(-1);
        }
     }

     if (tsnode->inode == NULL) {
        tsnode->inode = node_ptr = inode;
     } else {
       node_ptr->next = inode;
       node_ptr = inode;
     }
    } 

    va_end(tuple_list); 

    /* Attempt to find it in the tuple space */

    len = (len + hash) % SIZEOF_TS;
    pthread_mutex_lock(&_linda_tuplespace[len].lock);
    
    ptsptr = NULL;

    for (tsptr = _linda_tuplespace[len].anode; tsptr != NULL; 
         tsptr = tsptr->next) {

         if (match(tsnode->inode, tsptr->inode)) {
            if (tsptr->type == OUT) {
               bind_val(tsnode->inode, tsptr->inode);

               /* Delink the out tuple */
               if (ptsptr == NULL) {
                  _linda_tuplespace[len].anode = tsptr->next;
               } else {
                  ptsptr->next = tsptr->next;
               }

               pthread_mutex_unlock(&_linda_tuplespace[len].lock);
               return 1;
            } 
         }

         ptsptr = tsptr;
    }

    pthread_mutex_unlock(&_linda_tuplespace[len].lock);
    return 0;
}

int rdp (char *tuple_mask, ...) 
{ 
    va_list tuple_list; 
    char *mask_ptr; 
    int len, hash, i;
    struct _INODE_LIST *inode, *node_list, *node_ptr;
    struct _TS_NODE *tsnode, *ptsptr, *tsptr;
 
    len = (strlen(tuple_mask) / 2);
    hash = 0;
    
    tsnode = (struct _TS_NODE *)(malloc(sizeof(struct _TS_NODE)));
    tsnode->type = RD;
    tsnode->inode = NULL;

    node_ptr = NULL;

    va_start(tuple_list, tuple_mask); 

    for (mask_ptr = tuple_mask, i = 1; *mask_ptr; mask_ptr+=2, i++) 
    { 
     if (*mask_ptr == '%') {
         switch (mask_ptr[1]) 
         { 
         case 's': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.s = va_arg(tuple_list, char *); 
           inode->type_tag = CHARP_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 3 * i;
           break;
         case 'd': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.i = va_arg(tuple_list, int); 
           inode->type_tag = INT_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 5 * i;
           break;
         case 'f': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.f = va_arg(tuple_list, float); 
           inode->type_tag = FLOAT_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 7 * i;
           break;
         case 'c': 
           inode = 
             (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
           inode->node.c = va_arg(tuple_list, char); 
           inode->type_tag = CHAR_TYPE;
           inode->field_tag = D_FIELD;
           inode->next = NULL;
           hash += 11 * i;
           break;
         default:
           fprintf(stderr, "Unknown type of an out.\n");
           exit(-1);
         } 
     } else {
        if (*mask_ptr == '?') {
           switch (mask_ptr[1]) 
           { 
            case 's': 
              inode = 
                (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
              inode->node.s = va_arg(tuple_list, char *); 
              inode->type_tag = CHARP_TYPE;
              inode->field_tag = Q_FIELD;
              inode->next = NULL;
              hash += 3 * i;
              break;
            case 'd': 
              inode = 
                (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
              inode->node.ip = va_arg(tuple_list, int *); 
              inode->type_tag = INT_TYPE;
              inode->field_tag = Q_FIELD;
              inode->next = NULL;
              hash += 5 * i;
              break;
            case 'f': 
              inode = 
                (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
              inode->node.fp = va_arg(tuple_list, float *); 
              inode->type_tag = FLOAT_TYPE;
              inode->field_tag = Q_FIELD;
              inode->next = NULL;
              hash += 7 * i;
              break;
            case 'c': 
              inode = 
                (struct _INODE_LIST *)(malloc(sizeof(struct _INODE_LIST)));
              inode->node.cp = va_arg(tuple_list, char *); 
              inode->type_tag = CHAR_TYPE;
              inode->field_tag = Q_FIELD;
              inode->next = NULL;
              hash += 11 * i;
              break;
            default:
              fprintf(stderr, "Unknown type of an in.\n");
              exit(-1);
         }
        } else {
           fprintf(stderr, "Don't know what it is - in ?!");
           exit(-1);
        }
     }

     if (tsnode->inode == NULL) {
        tsnode->inode = node_ptr = inode;
     } else {
       node_ptr->next = inode;
       node_ptr = inode;
     }
    } 

    va_end(tuple_list); 

    /* Attempt to find it in the tuple space */

    len = (len + hash) % SIZEOF_TS;
    pthread_mutex_lock(&_linda_tuplespace[len].lock);
    
    ptsptr = NULL;

    for (tsptr = _linda_tuplespace[len].anode; tsptr != NULL; 
         tsptr = tsptr->next) {

         if (match(tsnode->inode, tsptr->inode)) {
            if (tsptr->type == OUT) {
               bind_val(tsnode->inode, tsptr->inode);

               /* Don't delink the out tuple */

               pthread_mutex_unlock(&_linda_tuplespace[len].lock);
               return 1;
            } 
         }

         ptsptr = tsptr;
    }

    pthread_mutex_unlock(&_linda_tuplespace[len].lock);
    return 0;
}

